"""Display-p3 color class."""
from ._space import RE_DEFAULT_MATCH
from .srgb import SRGB, lin_srgb, gam_srgb
from .xyz import XYZ
from . import _convert as convert
from .. import util
import re


def lin_p3_to_xyz(rgb):
    """
    Convert an array of linear-light image-p3 values to CIE XYZ using  D65 (no chromatic adaptation).

    http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
    """

    m = [
        [0.4865709486482162, 0.26566769316909306, 0.1982172852343625],
        [0.2289745640697488, 0.6917385218365064, 0.079286914093745],
        [0.0000000000000000, 0.04511338185890264, 1.043944368900976]
    ]

    # 0 was computed as -3.972075516933488e-17
    return util.dot(m, rgb)


def xyz_to_lin_p3(xyz):
    """Convert XYZ to linear-light P3."""

    m = [
        [2.493496911941425, -0.9313836179191239, -0.40271078445071684],
        [-0.8294889695615747, 1.7626640603183463, 0.023624685841943577],
        [0.03584583024378447, -0.07617238926804182, 0.9568845240076872]
    ]

    return util.dot(m, xyz)


def lin_p3(rgb):
    """Convert an array of image-p3 RGB values in the range 0.0 - 1.0 to linear light (un-corrected) form."""

    return lin_srgb(rgb)  # same as sRGB


def gam_p3(rgb):
    """Convert an array of linear-light image-p3 RGB  in the range 0.0-1.0 to gamma corrected form."""

    return gam_srgb(rgb)  # same as sRGB


class Display_P3(SRGB):
    """Display-p3 class."""

    SPACE = "display-p3"
    DEF_VALUE = "color(display-p3 0 0 0 / 1)"
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    WHITE = convert.WHITES["D65"]

    def __init__(self, color=DEF_VALUE):
        """Initialize."""

        super().__init__(color)

    @classmethod
    def _to_xyz(cls, rgb):
        """To XYZ."""

        return cls._chromatic_adaption(cls.white(), XYZ.white(), lin_p3_to_xyz(lin_p3(rgb)))

    @classmethod
    def _from_xyz(cls, xyz):
        """From XYZ."""

        return gam_p3(xyz_to_lin_p3(cls._chromatic_adaption(XYZ.white(), cls.white(), xyz)))



"""A98 RGB color class."""
from ._space import RE_DEFAULT_MATCH
from .srgb import SRGB
from .xyz import XYZ
from . import _convert as convert
from .. import util
import re
import math


def lin_a98rgb_to_xyz(rgb):
    """
    Convert an array of linear-light a98-rgb values to CIE XYZ using D50.D65.

    (so no chromatic adaptation needed afterwards)
    http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
    which has greater numerical precision than section 4.3.5.3 of
    https://www.adobe.com/digitalimag/pdfs/AdobeRGB1998.pdf
    """

    m = [
        [0.5766690429101305, 0.1855582379065463, 0.1882286462349947],
        [0.29734497525053605, 0.6273635662554661, 0.07529145849399788],
        [0.02703136138641234, 0.07068885253582723, 0.9913375368376388]
    ]

    return util.dot(m, rgb)


def xyz_to_lin_a98rgb(xyz):
    """Convert XYZ to linear-light a98-rgb."""

    m = [
        [2.0415879038107465, -0.5650069742788596, -0.34473135077832956],
        [-0.9692436362808795, 1.8759675015077202, 0.04155505740717557],
        [0.013444280632031142, -0.11836239223101838, 1.0151749943912054]
    ]

    return util.dot(m, xyz)


def lin_a98rgb(rgb):
    """Convert an array of a98-rgb values in the range 0.0 - 1.0 to linear light (un-corrected) form."""

    return [math.copysign(math.pow(abs(val), 563 / 256), val) for val in rgb]


def gam_a98rgb(rgb):
    """Convert an array of linear-light a98-rgb  in the range 0.0-1.0 to gamma corrected form."""

    return [math.copysign(math.pow(abs(val), 256 / 563), val) for val in rgb]


class A98_RGB(SRGB):
    """A98 RGB class."""

    SPACE = "a98-rgb"
    DEF_VALUE = "color(a98-rgb 0 0 0 / 1)"
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    WHITE = convert.WHITES["D65"]

    def __init__(self, color=DEF_VALUE):
        """Initialize."""

        super().__init__(color)

    @classmethod
    def _to_xyz(cls, rgb):
        """To XYZ."""

        return cls._chromatic_adaption(cls.white(), XYZ.white(), lin_a98rgb_to_xyz(lin_a98rgb(rgb)))

    @classmethod
    def _from_xyz(cls, xyz):
        """From XYZ."""

        return gam_a98rgb(xyz_to_lin_a98rgb(cls._chromatic_adaption(XYZ.white(), cls.white(), xyz)))



"""LAB class."""
from ._space import Space, RE_DEFAULT_MATCH
from ._gamut import GamutUnbound
from . _range import Percent
from . import _parse as parse
from . import _convert as convert
from .. import util
import re
import math

KAPPA = 24389 / 27  # `29^3 / 3^3`
EPSILON = 216 / 24389  # `6^3 / 29^3`


def lab_to_xyz(lab):
    """
    Convert Lab to D50-adapted XYZ.

    http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
    """

    l, a, b = lab

    # compute `f`, starting with the luminance-related term
    f1 = (l + 16) / 116
    f0 = a / 500 + f1
    f2 = f1 - b / 200

    # compute `xyz`
    xyz = [
        math.pow(f0, 3) if math.pow(f0, 3) > EPSILON else (116 * f0 - 16) / KAPPA,
        math.pow((l + 16) / 116, 3) if l > KAPPA * EPSILON else l / KAPPA,
        math.pow(f2, 3) if math.pow(f2, 3) > EPSILON else (116 * f2 - 16) / KAPPA
    ]

    # Compute XYZ by scaling `xyz` by reference `white`
    return util.multiply(xyz, LAB.white())


def xyz_to_lab(xyz):
    """Assuming XYZ is relative to D50, convert to CIE Lab from CIE standard."""

    # compute `xyz`, which is XYZ scaled relative to reference white
    xyz = util.divide(xyz, LAB.white())
    # Compute `f`
    f = [util.cbrt(i) if i > EPSILON else (KAPPA * i + 16.0) / 116.0 for i in xyz]

    return (
        (116.0 * f[1]) - 16.0,
        500.0 * (f[0] - f[1]),
        200.0 * (f[1] - f[2])
    )


class LAB(Space):
    """LAB class."""

    SPACE = "lab"
    DEF_VALUE = "color(lab 0 0 0 / 1)"
    CHANNEL_NAMES = frozenset(["lightness", "a", "b", "alpha"])
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    WHITE = convert.WHITES["D50"]

    _range = (
        GamutUnbound([Percent(0), Percent(100.0)]),  # Technically we could/should clamp the zero side.
        GamutUnbound([-160, 160]),  # No limit, but we could impose one +/-160?
        GamutUnbound([-160, 160])  # No limit, but we could impose one +/-160?
    )

    def __init__(self, color=DEF_VALUE):
        """Initialize."""

        super().__init__(color)

        if isinstance(color, Space):
            self.lightness, self.a, self.b = color.convert(self.space()).coords()
            self.alpha = color.alpha
        elif isinstance(color, str):
            values = self.match(color)[0]
            if values is None:
                raise ValueError("'{}' does not appear to be a valid color".format(color))
            self.lightness, self.a, self.b, self.alpha = values
        elif isinstance(color, (list, tuple)):
            if not (3 <= len(color) <= 4):
                raise ValueError("A list of channel values should be of length 3 or 4.")
            self.lightness = color[0]
            self.a = color[1]
            self.b = color[2]
            self.alpha = 1.0 if len(color) == 3 else color[3]
        else:
            raise TypeError("Unexpected type '{}' received".format(type(color)))

    @property
    def lightness(self):
        """L channel."""

        return self._coords[0]

    @lightness.setter
    def lightness(self, value):
        """Get true luminance."""

        self._coords[0] = self._handle_input(value)

    @property
    def a(self):
        """A channel."""

        return self._coords[1]

    @a.setter
    def a(self, value):
        """A axis."""

        self._coords[1] = self._handle_input(value)

    @property
    def b(self):
        """B channel."""

        return self._coords[2]

    @b.setter
    def b(self, value):
        """B axis."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def translate_channel(cls, channel, value):
        """Translate channel string."""

        if 0 <= channel <= 2:
            return parse.norm_float(value)
        elif channel == -1:
            return parse.norm_alpha_channel(value)
        else:
            raise ValueError("Unexpected channel index of '{}'".format(channel))

    @classmethod
    def _to_xyz(cls, lab):
        """To XYZ."""

        return lab_to_xyz(lab)

    @classmethod
    def _from_xyz(cls, xyz):
        """From XYZ."""

        return xyz_to_lab(xyz)



"""Parse utilities."""
import re
import math
from .. import util

RGB_CHANNEL_SCALE = 1.0 / 255.0
HUE_SCALE = 1.0 / 360.0
SCALE_PERCENT = 1 / 100.0

CONVERT_TURN = 360
CONVERT_GRAD = 90 / 100

RE_BRACKETS = re.compile(r'(?:(\()|(\))|[^()]+)')
RE_CHAN_SPLIT = re.compile(r'(?:\s*[,/]\s*|\s+)')
RE_COMMA_SPlIT = re.compile(r'(?:\s*,\s*)')
RE_SLASH_SPLIT = re.compile(r'(?:\s*/\s*)')

COLOR_PARTS = {
    "percent": r"[+\-]?(?:(?:[0-9]*\.[0-9]+)|[0-9]+)(?:e[-+]?[0-9]*)?%",
    "float": r"[+\-]?(?:(?:[0-9]*\.[0-9]+)|[0-9]+)(?:e[-+]?[0-9]*)?",
    "angle": r"[+\-]?(?:(?:[0-9]*\.[0-9]+)|[0-9]+)(?:e[-+]?[0-9]*)?(deg|rad|turn|grad)?",
    "space": r"\s+",
    "comma": r"\s*,\s*",
    "slash": r"\s*/\s*",
    "sep": r"(?:\s*,\s*|\s+)",
    "asep": r"(?:\s*[,/]\s*|\s+)",
    "hex": r"[a-f0-9]"
}


def norm_float(string):
    """Normalize a float value."""

    if string.lower().endswith(('e-', 'e+', 'e')):
        string += '0'
    return float(string)


def norm_hex_channel(string):
    """Normalize the hex string to a form we can handle."""

    if string.startswith('#'):
        return int(string[1:], 16) * RGB_CHANNEL_SCALE
    else:
        raise ValueError("Unexpected value '{}'".format(string))


def norm_percent_channel(value, scale=False):
    """Normalize percent channel."""

    if value.endswith('%'):
        value = norm_float(value[:-1])
        return value / 100.0 if scale else value
    else:
        raise ValueError("Unexpected value '{}'".format(value))


def norm_color_channel(value, scale=True):
    """Normalize percent channel."""

    if value.endswith('%'):
        value = norm_float(value[:-1])
        return value / 100.0 if scale else value
    else:
        return norm_float(value)


def norm_rgb_channel(value):
    """Normalize RGB channel."""

    if value.endswith("%"):
        return norm_percent_channel(value, True)
    else:
        return norm_float(value) * RGB_CHANNEL_SCALE


def norm_alpha_channel(value):
    """Normalize alpha channel."""

    if value.endswith("%"):
        value = norm_percent_channel(value, True)
    else:
        value = norm_float(value)
    return util.clamp(value, 0.0, 1.0)


def norm_lab_lightness(value):
    """Normalize lab channel."""

    if value.endswith('%'):
        return norm_float(value[:-1])
    else:
        raise ValueError("Unexpected value '{}'".format(value))


def norm_angle(angle):
    """Normalize angle units."""

    if angle.endswith('turn'):
        value = norm_float(angle[:-4]) * CONVERT_TURN
    elif angle.endswith('grad'):
        value = norm_float(angle[:-4]) * CONVERT_GRAD
    elif angle.endswith('rad'):
        value = math.degrees(norm_float(angle[:-3]))
    elif angle.endswith('grad'):
        value = norm_float(angle[:-3]) * CONVERT_GRAD
    elif angle.endswith('deg'):
        value = norm_float(angle[:-3])
    else:
        value = norm_float(angle)
    return value


def norm_hue_channel(value):
    """Normalize hex channel."""

    return norm_angle(value)


def norm_deg_channel(value):
    """Normalize degree channel."""

    return norm_float(value)


def bracket_match(match, string, start, fullmatch):
    """
    Make sure we can acquire a complete `func()` before we replace variables.

    We mainly do this so we can judge the real size before we alter the string with variables.
    """

    end = None
    if match.match(string, start):
        brackets = 1
        for m in RE_BRACKETS.finditer(string, start + 6):
            if m.group(2):
                brackets -= 1
            elif m.group(1):
                brackets += 1

            if brackets == 0:
                end = m.end(2)
                break
    return end if (not fullmatch or end == len(string)) else None



"""Gamut handling."""
from .. import util
from . _range import Angle


class GamutBound(tuple):
    """Bounded gamut value."""


class GamutUnbound(tuple):
    """Unbounded gamut value."""


def lch_chroma(base, color):
    """
    Gamut mapping via chroma LCH.

    Algorithm comes from https://colorjs.io/docs/gamut-mapping.html.

    The idea is to hold hue and lightness constant and decrease lightness until
    color comes under gamut.

    We'll use a binary search and at after each stage, we will clip the color
    and compare the distance of the two colors (clipped and current color via binary search).
    If the distance is less than two, we can return the clipped color.

    ---
    Original Authors: Lea Verou, Chris Lilley
    License: MIT (As noted in https://github.com/LeaVerou/color.js/blob/master/package.json)
    """

    # Compare clipped against original to
    # judge how far we are off with the worst case fitting
    space = color.space()
    clipped = color.clone()
    clipped.fit(space=space, method="clip", in_place=True)
    base_error = base.delta_e(clipped, method="2000")

    if base_error > 2.3:
        threshold = .001
        # Compare mapped against desired space
        mapcolor = color.convert("lch")
        error = color.delta_e(mapcolor, method="2000")
        low = 0.0
        high = mapcolor.chroma

        # Adjust chroma (using binary search).
        # This helps preserve the color more (in most cases).
        # After each adjustment, see if clipping gets us close enough.
        while (high - low) > threshold and error < base_error:
            clipped = mapcolor.clone()
            clipped.fit(space, method="clip", in_place=True)
            delta = mapcolor.delta_e(clipped, method="2000")
            error = color.delta_e(mapcolor, method="2000")
            if delta - 2 < threshold:
                low = mapcolor.chroma
            else:
                if abs(delta - 2) < threshold:
                    break
                high = mapcolor.chroma
            mapcolor.chroma = (high + low) / 2
        # Trim off noise allowed by our tolerance
        color.update(mapcolor)
        color.fit(space, method="clip", in_place=True)
    else:
        # We are close enough that we should just clip.
        color.update(clipped)
    return color.coords()


def clip(base, color):
    """Gamut clipping."""

    channels = util.no_nan(color.coords())
    gamut = color._range
    fit = []

    for i, value in enumerate(channels):
        a, b = gamut[i]
        is_bound = isinstance(gamut[i], GamutBound)

        # Wrap the angle. Not technically out of gamut, but we will clean it up.
        if isinstance(a, Angle) and isinstance(b, Angle):
            fit.append(value % 360.0)
            continue

        # These parameters are unbounded
        if not is_bound:
            a = None
            b = None

        # Fit value in bounds.
        fit.append(util.clamp(value, a, b))
    return fit


def norm_angles(color):
    """Normalize angles."""

    channels = util.no_nan(color.coords())
    gamut = color._range
    fit = []
    for i, value in enumerate(channels):
        a = gamut[i][0]

        # Wrap the angle
        if isinstance(a, Angle):
            fit.append(value % 360.0)
            continue

        # Fit value in bounds.
        fit.append(value)
    return fit


class Gamut:
    """Gamut handling."""

    def fit_coords(self, space=None, *, method=None):
        """Get coordinates within this space or fit to another space."""

        if method is None:
            method = self.parent.FIT

        space = (self.space() if space is None else space).lower()
        method = self.space() if method is None else method
        clone = self.clone()
        if not clone.in_gamut(space=space, tolerance=0.0):
            clone.fit(method=method, in_place=True)
            return clone.coords()
        else:
            return norm_angles(clone)

    def fit(self, space=None, *, method=None, in_place=False):
        """Fit the gamut using the provided method."""

        if method is None:
            method = self.parent.FIT

        this = self if in_place else self.clone()

        # Select appropriate mapping algorithm
        if method == "clip":
            func = clip
        elif method == "lch-chroma":
            func = lch_chroma
        else:
            # Unknown fit method
            raise ValueError("'{}' gamut mapping is not currently supported".format(method))

        # Convert to desired space
        if space is not None:
            c = self.convert(space)
        else:
            c = self.clone()

        # If we are perfectly in gamut, don't waste time fitting
        if c.in_gamut(tolerance=0.0):
            c._coords = norm_angles(c)
            this.update(c)
            return this

        # Apply mapping/clipping/etc.
        c._coords = func(self.clone(), c)

        # Adjust "this" color
        this.update(c)
        return this

    def in_gamut(self, space=None, *, tolerance=util.DEF_FIT_TOLERANCE):
        """Check if current color is in gamut."""

        # Check gamut in the provided space
        if space is not None:
            c = self.convert(space)
            return c.in_gamut(tolerance=tolerance)

        # Check the color space specified for gamut checking.
        # If it proves to be in gamut, we will then test if the current
        # space is constrained properly.
        if self.GAMUT_CHECK is not None:  # pragma: no cover
            c2 = self.convert(self.GAMUT_CHECK)
            if not c2.in_gamut(tolerance=tolerance):
                return False

        # Verify the values are in bound
        channels = util.no_nan(self.coords())
        for i, value in enumerate(channels):
            a, b = self._range[i]
            is_bound = isinstance(self._range[i], GamutBound)

            # Angles will wrap, so no sense checking them
            if isinstance(a, Angle):
                continue

            # These parameters are unbounded
            if not is_bound:
                a = None
                b = None

            # Check if bounded values are in bounds
            if (a is not None and value < (a - tolerance)) or (b is not None and value > (b + tolerance)):
                return False

        return True



"""
Interpolation methods.

A lot of code was ported and or adapted from the https://colorjs.io project. Particularly
the `interpolate` method and the functions built on top of it, such as `mix` and `steps`.

While we deviate in some ways, a lot of it, at the time of this comment, are a direct port.

In general, the logic mimics in many ways the `color-mix` function as outlined in the Level 5
color draft (Oct 2020), but the approach was modeled directly off of the work done in color.js.
---
Original Authors: Lea Verou, Chris Lilley
License: MIT (As noted in https://github.com/LeaVerou/color.js/blob/master/package.json)
"""
import math
import functools
from .. import util
from . _cylindrical import Cylindrical
from . _range import Angle


def overlay(c1, c2, a1, a2, a0, angle=False):
    """Overlay one color channel over the other."""

    if util.is_nan(c1) and util.is_nan(c2):
        return 0.0
    elif util.is_nan(c1):
        return c2 if angle else c2 * a2
    elif util.is_nan(c2):
        return c1 if angle else c1 * a1

    if angle:
        return c1 + (c2 - c1) * (1 - a1)
    else:
        c0 = c1 * a1 + c2 * a2 * (1 - a1)
        return c0 / a0 if a0 else c0


def interpolate(p, coords1, coords2, create, progress, outspace, premultiplied):
    """Run through the coordinates and run the interpolation on them."""

    coords = []
    for i, c1 in enumerate(coords1):
        c2 = coords2[i]
        if util.is_nan(c1) and util.is_nan(c2):
            value = 0.0
        elif util.is_nan(c1):
            value = c2
        elif util.is_nan(c2):
            value = c1
        else:
            value = c1 + (c2 - c1) * (p if progress is None else progress(p))
        coords.append(value)
    color = create.new(coords).convert(outspace)
    if premultiplied:
        postdivide(color)
    return color


def prepare_coords(color, adjust=None):
    """
    Prepare the coordinates for interpolation.

    If the hue is null, we need to set it to NaN.
    If the user specified only specific channels to mix,
    then we need to set all other channels to NaN.
    """

    if adjust:
        to_adjust = adjust & color.CHANNEL_NAMES
        to_avoid = color.CHANNEL_NAMES - adjust
        if to_adjust:
            for channel in to_avoid:
                color.set(channel, util.NaN)


def postdivide(color):
    """Premultiply the given transparent color."""

    if color.alpha >= 1.0:
        return

    channels = color.coords()
    gamut = color._range
    alpha = color.alpha
    coords = []
    for i, value in enumerate(channels):
        a = gamut[i][0]

        # Wrap the angle
        if isinstance(a, Angle):
            coords.append(value)
            continue
        coords.append(value / alpha if alpha != 0 else value)
    color._coords = coords


def premultiply(color):
    """Premultiply the given transparent color."""

    if color.alpha >= 1.0:
        return

    channels = color.coords()
    gamut = color._range
    alpha = color.alpha
    coords = []
    for i, value in enumerate(channels):
        a = gamut[i][0]

        # Wrap the angle
        if isinstance(a, Angle):
            coords.append(value)
            continue
        coords.append(value * alpha)
    color._coords = coords


def adjust_hues(color1, color2, hue):
    """Adjust hues."""

    hue = hue.lower()
    if hue == "specified":
        return

    name = color1.hue_name()
    c1 = color1.get(name)
    c2 = color2.get(name)

    c1 = c1 % 360
    c2 = c2 % 360

    if util.is_nan(c1) or util.is_nan(c2):
        color1.set(name, c1)
        color2.set(name, c2)
        return

    if hue == "shorter":
        if c2 - c1 > 180:
            c1 += 360
        elif c2 - c1 < -180:
            c2 += 360

    elif hue == "longer":
        if 0 < (c2 - c1) < 180:
            c1 += 360
        elif -180 < (c2 - c1) < 0:
            c2 += 360

    elif hue == "increasing":
        if c2 < c1:
            c2 += 360

    elif hue == "decreasing":
        if c1 < c2:
            c1 += 360

    else:
        raise ValueError("Unknown hue adjuster '{}'".format(hue))

    color1.set(name, c1)
    color2.set(name, c2)


class Interpolate:
    """Interpolate between colors."""

    def overlay(self, background, *, space=None, in_place=False):
        """
        Apply the given transparency with the given background.

        This attempts to give a color that represents what the eye
        sees with the transparent color against the given background.

        Some spaces will require the action to take place in a different
        space. For instance, cylindrical representations of sRGB (HSL, HSV, HWB),
        will request interpolation to be done under sRGB. This is because alpha
        compositing does not work well in cylindrical spaces.
        """

        current_space = self.space()
        if self.alpha < 1.0:
            if space is None:
                space = current_space
            else:
                space = space.lower()

            this = self.convert(space, fit=True)
            background = background.convert(space, fit=True)

            if this is None:
                raise ValueError('Invalid colorspace value: {}'.format(space))

            # Get the coordinates and indexes of valid hues
            prepare_coords(this)
            prepare_coords(background)

            # Adjust hues if we have two valid hues
            if isinstance(this, Cylindrical):
                adjust_hues(this, background, util.DEF_HUE_ADJ)

            coords1 = this.coords()
            coords2 = background.coords()
            a1 = this.alpha
            a2 = background.alpha
            a0 = a1 + a2 * (1.0 - a1)
            gamut = this._range
            coords = []
            # Avoid multiplying angles and don't mix them the same as non-angles
            for i, value in enumerate(coords1):
                g = gamut[i][0]
                is_angle = isinstance(g, Angle)
                coords.append(overlay(coords1[i], coords2[i], a1, a2, a0, angle=is_angle))
            this._coords = coords
            this.alpha = a0
        else:
            this = self

        if in_place:
            return self.update(this.convert(current_space))

        return this.convert(current_space)

    def steps(self, color, *, steps=2, max_steps=1000, max_delta_e=0, **interpolate_args):
        """
        Discrete steps.

        This is built upon the interpolate function, and will return a list of
        colors containing a minimum of colors equal to `steps` or steps as specified
        derived from the `max_delta_e` parameter (whichever is greatest).

        Number of colors can be capped with `max_steps`.

        Default delta E method used is delta E 76.
        """

        interp = self.interpolate(color, **interpolate_args)
        total_delta = self.delta_e(color)
        actual_steps = steps if max_delta_e <= 0 else max(steps, math.ceil(total_delta / max_delta_e) + 1)
        if max_steps is not None:
            actual_steps = min(actual_steps, max_steps)

        ret = []
        if actual_steps == 1:
            ret = [{"p": 0.5, "color": interp(0.5)}]
        else:
            step = 1 / (actual_steps - 1)
            for i in range(actual_steps):
                p = i * step
                ret.append({'p': p, 'color': interp(p)})

        # Iterate over all the stops inserting stops in between if all colors
        # if we have any two colors with a max delta greater than what was requested.
        # We inject between every stop to ensure the midpoint does not shift.
        if max_delta_e > 0:
            # Initial check to see if we need to insert more stops
            m_delta = 0
            for i, entry in enumerate(ret):
                if i == 0:
                    continue
                m_delta = max(m_delta, entry['color'].delta_e(ret[i - 1]['color']))

            while m_delta > max_delta_e:
                # Inject stops while measuring again to see if it was sufficient
                m_delta = 0
                i = 1
                while i < len(ret) and len(ret) < max_steps:
                    prev = ret[i - 1]
                    cur = ret[i]
                    p = (cur['p'] + prev['p']) / 2
                    color = interp(p)
                    m_delta = max(m_delta, color.delta_e(prev['color']), color.delta_e(cur['color']))
                    ret.insert(i, {'p': p, 'color': color})
                    i += 2

        return [i['color'] for i in ret]

    def mix(self, color, percent=util.DEF_MIX, *, space=None, in_place=False, **interpolate_args):
        """
        Mix colors using interpolation.

        This uses the interpolate method to find the center point between the two colors.
        The basic mixing logic is outlined in the CSS level 5 draft.
        """

        if space is None:
            space = self.space()
        else:
            space = space.lower()

        obj = self.interpolate(color, space=space, **interpolate_args)(percent)
        if in_place:
            return self.update(obj)
        return obj

    def interpolate(
        self, color, *, space="lab", out_space=None, progress=None, adjust=None, hue=util.DEF_HUE_ADJ,
        premultiplied=False
    ):
        """
        Return an interpolation function.

        The function will return an interpolation function that accepts a value (which should
        be in the range of [0..1] and will return a color based on that value.

        While we use NaNs to mask off channels when doing the interpolation, we do not allow
        arbitrary specification of NaNs by the user, they must specify channels via `adjust`
        if they which to target specific channels for mixing. Null hues become NaNs before
        mixing occurs.
        """

        if progress is not None and not callable(progress):
            raise TypeError('Progress must be callable')

        if adjust:
            adjust = set([name.lower() for name in adjust])

        inspace = space.lower()
        outspace = self.space() if out_space is None else out_space

        # Convert to the color space and ensure the color fits inside
        color1 = self.convert(inspace, fit=True)
        color2 = color.convert(inspace, fit=True)

        # Get the coordinates and indexes of valid hues
        prepare_coords(color1)
        prepare_coords(color2, adjust)

        # Adjust hues if we have two valid hues
        if isinstance(color1, Cylindrical):
            adjust_hues(color1, color2, hue)

        if premultiplied:
            premultiply(color1)
            premultiply(color2)

        coords1 = color1.coords()
        coords2 = color2.coords()

        # Include alpha
        coords1.append(color1.alpha)
        coords2.append(color2.alpha)

        return functools.partial(
            interpolate,
            coords1=coords1,
            coords2=coords2,
            create=color1,
            progress=progress,
            outspace=outspace,
            premultiplied=premultiplied
        )



"""Test API."""
import unittest
from coloraide.css import Color


class TestAPI(unittest.TestCase):
    """Test API."""

    def test_new(self):
        """Test new."""

        c1 = Color('purple')
        c2 = c1.new('purple')
        self.assertEqual(c1, c2)

    def test_clone(self):
        """Test clone."""

        c1 = Color('purple')
        c2 = c1.clone()
        self.assertEqual(c1, c2)

    def test_update(self):
        """Test update."""

        c1 = Color('orange')
        c2 = Color('purple')
        self.assertNotEqual(c1, c2)
        c2.update(c1)
        self.assertEqual(c1, c2)

    def test_update_different_space(self):
        """Update from different space."""

        c1 = Color('orange')
        c2 = c1.convert('hsl')
        c3 = Color('purple').convert('hsl')
        self.assertEqual(c2, c3.update(c1))

    def test_mutate(self):
        """Test mutate."""

        c1 = Color('orange').convert('lch')
        c2 = Color('orange')
        self.assertNotEqual(c1, c2)
        c2.mutate(c1)
        self.assertEqual(c1, c2)

    def test_luminance(self):
        """Test luminance."""

        c1 = Color('orange')
        c2 = c1.convert('xyz')
        self.assertEqual(c1.luminance(), c2.y)

    def test_property(self):
        """Test set."""

        c1 = Color('red')
        self.assertEqual(c1.green, 0)
        c1.green = 0.5
        self.assertEqual(c1.green, 0.5)

    def test_get(self):
        """Test get."""

        c1 = Color('orange')
        self.assertEqual(c1.get("red"), 1.0)

    def test_space_get(self):
        """Test get with another space."""

        c1 = Color('orange')
        self.assertEqual(c1.get("hsl.lightness"), 50.0)

    def test_set(self):
        """Test get."""

        c1 = Color('orange')
        c1.set("red", 0.5)
        self.assertEqual(c1.get("red"), 0.5)

    def test_space_set(self):
        """Test set in another space."""

        c1 = Color('orange')
        c1.set("hsl.hue", 270)
        self.assertEqual(c1.get("hsl.hue"), 270)

    def test_function_set(self):
        """Test set with a function."""

        c1 = Color('orange')
        c1.set("red", lambda x: x * 0.3)
        self.assertEqual(c1.get("red"), 0.3)

    def test_overlay(self):
        """Test overlay logic."""

        c1 = Color('blue').set('alpha', 0.5)
        c2 = Color('yellow')
        self.assertEqual(c1.overlay(c2), Color('color(srgb 0.5 0.5 0.5)'))

    def test_overlay_cyl(self):
        """Test overlay logic."""

        c1 = Color('blue').set('alpha', 0.5)
        c2 = Color('yellow')
        self.assertEqual(c1.overlay(c2, space="hsl"), Color('color(srgb 0 1 0.5)'))



"""Colors."""
from .hsv import HSV
from .srgb import SRGB
from .srgb_linear import SRGB_Linear
from .hsl import HSL
from .hwb import HWB
from .lab import LAB
from .lch import LCH
from .display_p3 import Display_P3
from .a98_rgb import A98_RGB
from .prophoto_rgb import ProPhoto_RGB
from .rec2020 import Rec2020
from .xyz import XYZ
from .. import util
import functools

DEF_FIT = "lch-chroma"
DEF_DELTA_E = "76"

SUPPORTED = (
    HSL, HWB, LAB, LCH, SRGB, SRGB_Linear, HSV,
    Display_P3, A98_RGB, ProPhoto_RGB, Rec2020, XYZ
)


def _interpolate(percent, color=None, interp=None):
    """Wrapper for interpolate."""

    obj = interp(percent)
    return color.new(obj.space(), obj.coords(), obj.alpha)


class ColorMatch:
    """Color match object."""

    def __init__(self, color, start, end):
        """Initialize."""

        self.color = color
        self.start = start
        self.end = end

    def __str__(self):
        """String."""

        return "ColorMatch(color={!r}, start={}, end={})".format(self.color, self.start, self.end)

    __repr__ = __str__


class Color:
    """Color class object which provides access and manipulation of color spaces."""

    CS_MAP = {obj.space(): obj for obj in SUPPORTED}

    PRECISION = util.DEF_PREC
    FIT = util.DEF_FIT
    DELTA_E = util.DEF_DELTA_E

    def __init__(self, color, data=None, alpha=util.DEF_ALPHA, *, filters=None, **kwargs):
        """Initialize."""

        self._attach(self._parse(color, data, alpha, filters=filters, **kwargs))

    def __eq__(self, other):
        """Compare equal."""

        return (
            other.space() == self.space() and
            util.cmp_coords(other.coords(), self.coords()) and
            util.cmp_coords(other.alpha, self.alpha)
        )

    def is_nan(self, name):
        """Check if channel is NaN."""

        return util.is_nan(self.get(name))

    def _attach(self, color):
        """Attach the this objects convert space to the color."""

        self._color = color
        self._color.parent = self

    def _handle_color_input(self, color):
        """Handle color input."""

        if isinstance(color, Color):
            color = color._color
        elif isinstance(color, str):
            color = self.new(color)._color
        else:
            raise TypeError("Unexpected type '{}'".format(type(color)))
        return color

    def _parse(self, color, data=None, alpha=util.DEF_ALPHA, filters=None, **kwargs):
        """Parse the color."""

        obj = None
        if data is not None:
            filters = set(filters) if filters is not None else set()
            for space, space_class in self.CS_MAP.items():
                s = color.lower()
                if space == s and (not filters or s in filters):
                    obj = space_class(data[:space_class.NUM_COLOR_CHANNELS] + [alpha])
                    return obj
        elif isinstance(color, Color):
            if not filters or color.space() in filters:
                obj = self.CS_MAP[color.space()](color._color)
        else:
            m = self._match(color, fullmatch=True, filters=filters)
            if m is None:
                raise ValueError("'{}' is not a valid color".format(color))
            obj = m.color
        if obj is None:
            raise ValueError("Could not process the provided color")
        return obj

    @classmethod
    def _match(cls, string, start=0, fullmatch=False, filters=None):
        """
        Match a color in a buffer and return a color object.

        This must return the color space, not the Color object.
        """

        filters = set(filters) if filters is not None else set()

        for space, space_class in cls.CS_MAP.items():
            if filters and space not in filters:
                continue
            value, match_end = space_class.match(string, start, fullmatch)
            if value is not None:
                color = space_class(value)
                return ColorMatch(color, start, match_end)
        return None

    @classmethod
    def match(cls, string, start=0, fullmatch=False, *, filters=None):
        """Match color."""

        obj = cls._match(string, start, fullmatch, filters=filters)
        if obj is not None:
            obj.color = cls(obj.color.space(), obj.color.coords(), obj.color.alpha)
        return obj

    def space(self):
        """The current color space."""

        return self._color.space()

    def coords(self):
        """Coordinates."""

        return self._color.coords()

    @classmethod
    def new(cls, color, data=None, alpha=util.DEF_ALPHA, *, filters=None, **kwargs):
        """Create new color object."""

        return cls(color, data, alpha, filters=filters, **kwargs)

    def clone(self):
        """Clone."""

        clone = self._color.clone()
        return self.new(clone.space(), clone.coords(), clone.alpha)

    def convert(self, space, *, fit=False, in_place=False):
        """Convert."""

        obj = self._color.convert(space, fit=fit)
        if in_place:
            return self._attach(obj)
        return type(self)(obj.space(), obj.coords(), obj.alpha)

    def update(self, color, data=None, alpha=util.DEF_ALPHA, *, filters=None, **kwargs):
        """Update the existing color space with the provided color."""

        clone = self.clone()
        obj = self._parse(color, data, alpha, filters=filters, **kwargs)
        clone._attach(obj)
        self._color.update(obj)
        return self

    def mutate(self, color, data=None, alpha=util.DEF_ALPHA, *, filters=None, **kwargs):
        """Mutate the current color to a new color."""

        self._attach(self._parse(color, data, alpha, filters=filters, **kwargs))
        return self

    def to_string(self, **kwargs):
        """To string."""

        return self._color.to_string(**kwargs)

    def __repr__(self):
        """Representation."""

        return repr(self._color)

    __str__ = __repr__

    def luminance(self):
        """Get color's luminance."""

        return self._color.luminance()

    def contrast(self, color):
        """Compare the contrast ratio of this color and the provided color."""

        color = self._handle_color_input(color)
        return self._color.contrast(color)

    def distance(self, color, *, space=util.DEF_DISTANCE_SPACE):
        """Get distance between this color and the provided color."""

        color = self._handle_color_input(color)
        return self._color.distance(color, space=space)

    def delta_e(self, color, *, method=None, **kwargs):
        """Delta E distance."""

        color = self._handle_color_input(color)
        return self._color.delta_e(color, method=method, **kwargs)

    def overlay(self, background=None, *, space=None, in_place=False):
        """Apply the given transparency with the given background."""

        background = self._handle_color_input(background)
        obj = self._color.overlay(background, space=space, in_place=in_place)

        if not in_place:
            return self.new(obj.space(), obj.coords(), obj.alpha)
        return self

    def interpolate(
        self, color, *, space="lab", out_space=None, progress=None, adjust=None, hue=util.DEF_HUE_ADJ,
        premultiplied=False
    ):
        """Interpolate."""

        color = self._handle_color_input(color)
        interp = self._color.interpolate(
            color, space=space, progress=progress, out_space=None, adjust=adjust, hue=hue, premultiplied=premultiplied
        )
        return functools.partial(_interpolate, color=self.clone(), interp=interp)

    def steps(self, color, *, steps=2, max_steps=1000, max_delta_e=0, **interpolate_args):
        """Interpolate discrete steps."""

        color = self._handle_color_input(color)
        colors = []
        for obj in self._color.steps(
            color, steps=steps, max_steps=max_steps, max_delta_e=max_delta_e, **interpolate_args
        ):
            colors.append(self.new(obj.space(), obj.coords(), obj.alpha))
        return colors

    def mix(self, color, percent=util.DEF_MIX, *, space=None, in_place=False, **interpolate_args):
        """Mix the two colors."""

        color = self._handle_color_input(color)
        obj = self._color.mix(color, percent, space=space, in_place=in_place, **interpolate_args)
        if not in_place:
            return self.new(obj.space(), obj.coords(), obj.alpha)
        return self

    def fit(self, space=None, *, method=None, in_place=False):
        """Fit to gamut."""

        obj = self._color.fit(space, method=method, in_place=in_place)
        if not in_place:
            return self.new(obj.space(), obj.coords(), obj.alpha)
        return self

    def in_gamut(self, space=None, *, tolerance=util.DEF_FIT_TOLERANCE):
        """Check if in gamut."""

        return self._color.in_gamut(space, tolerance=tolerance)

    def get(self, name):
        """Get channel."""

        # Handle space.attribute
        if '.' in name:
            parts = name.split('.')
            if len(parts) != 2:
                raise ValueError("Could not resolve attribute '{}'".format(name))
            obj = self.convert(parts[0])
            return obj.get(parts[1])

        return self._color.get(name)

    def set(self, name, value):  # noqa: A003
        """Set channel."""

        # Handle space.attribute
        if '.' in name:
            parts = name.split('.')
            if len(parts) != 2:
                raise ValueError("Could not resolve attribute '{}'".format(name))
            obj = self.convert(parts[0])
            obj.set(parts[1], value)
            return self.update(obj)

        # Handle a function that modifies the value or a direct value
        if callable(value):
            self.set(name, value(self.get(name)))
        else:
            self._color.set(name, value)
        return self

    def __getattr__(self, name):
        """Get attribute."""

        # Don't test `_color` as it is used to get Space channel attributes.
        if name != "_color":
            # Get channel names
            names = set()
            result = getattr(self, "_color")
            if result is not None:
                names = result.CHANNEL_NAMES
            # If requested attribute is a channel name, return the attribute from the Space instance.
            if name in names:
                return getattr(result, name)

    def __setattr__(self, name, value):
        """Set attribute."""

        try:
            # See if we need to set the space specific channel attributes.
            if name in self._color.CHANNEL_NAMES:
                setattr(self._color, name, value)
                return
        except AttributeError:
            pass
        # Set all attributes on the Color class.
        super().__setattr__(name, value)



"""Test SRGB library."""
import unittest
from coloraide.css import Color


class TestSRGBParse(unittest.TestCase):
    """Test SRGB."""

    def test_comma_comma(self):
        """Test comma format."""

        args = {"comma": True, "precision": 0}

        color = "rgb(30, 100, 200)"
        rgb = Color(color)
        self.assertEqual(color, rgb.to_string(**args))

        color = "rgb(30, 100, 200, 1)"
        rgb = Color(color)
        self.assertEqual("rgb(30, 100, 200)", rgb.to_string(**args))

        color = "rgb(30, 100, 200, 0.2)"
        rgb = Color(color)
        self.assertEqual("rgba(30, 100, 200, 0.2)", rgb.to_string(**args))

    def test_space_space(self):
        """Test space format."""

        args = {"precision": 0}

        color = "rgb(30 100 200)"
        rgb = Color(color)
        self.assertEqual(color, rgb.to_string(**args))

        color = "rgb(30 100 200 / 1)"
        rgb = Color(color)
        self.assertEqual("rgb(30 100 200)", rgb.to_string(**args))

        color = "rgb(30 100 200 / 0.2)"
        rgb = Color(color)
        self.assertEqual(color, rgb.to_string(**args))

    def test_hex_hex(self):
        """Test hex input."""

        args = {"hex": True, "precision": 0}

        color = "#AA33DD"
        rgb = Color(color)
        self.assertEqual(color.lower(), rgb.to_string(**args))

        color = "#AA33DDFF"
        rgb = Color(color)
        self.assertEqual("#aa33dd", rgb.to_string(**args))

        color = "#AA33DD22"
        rgb = Color(color)
        self.assertEqual("#aa33dd22", rgb.to_string(**args))

    def test_hex_compressed(self):
        """Test hex input."""

        args = {"hex": True, "compress": True}

        color = "#AA33DD"
        rgb = Color(color)
        self.assertEqual("#a3d", rgb.to_string(**args))

        color = "#AA33DDFF"
        rgb = Color(color)
        self.assertEqual("#a3d", rgb.to_string(**args))

        args["alpha"] = True

        color = "#AA33DD22"
        rgb = Color(color)
        self.assertEqual("#a3d2", rgb.to_string(**args))

    def test_alias(self):
        """Test that `rgba` is an alias for `rgb`."""

        args = {"comma": True, "precision": 0}

        color = "rgba(30, 100, 200)"
        rgb = Color(color)
        self.assertEqual("rgb(30, 100, 200)", rgb.to_string(**args))

        color = "rgba(30, 100, 200, 1)"
        rgb = Color(color)
        self.assertEqual("rgb(30, 100, 200)", rgb.to_string(**args))

        args["comma"] = False

        color = "rgba(30 100 200)"
        rgb = Color(color)
        self.assertEqual("rgb(30 100 200)", rgb.to_string(**args))

        color = "rgba(30 100 200 / 1)"
        rgb = Color(color)
        self.assertEqual("rgb(30 100 200)", rgb.to_string(**args))

    def test_percent(self):
        """Test that `rgba` is an alias for `rgb`."""

        args = {"comma": True, "precision": 0}

        color = "rgb(30%, 100%, 100%)"
        rgb = Color(color)
        self.assertEqual("rgb(77, 255, 255)", rgb.to_string(**args))

        color = "rgba(30%, 100%, 100%, 100%)"
        rgb = Color(color)
        self.assertEqual("rgb(77, 255, 255)", rgb.to_string(**args))

        color = "rgba(30%, 100%, 100%, 20%)"
        rgb = Color(color)
        self.assertEqual("rgba(77, 255, 255, 0.2)", rgb.to_string(**args))

        args["comma"] = False

        color = "rgb(30% 100% 100%)"
        rgb = Color(color)
        self.assertEqual("rgb(77 255 255)", rgb.to_string(**args))

        color = "rgba(30% 100% 100% / 100%)"
        rgb = Color(color)
        self.assertEqual("rgb(77 255 255)", rgb.to_string(**args))

        color = "rgba(30% 100% 100% / 20%)"
        rgb = Color(color)
        self.assertEqual("rgb(77 255 255 / 0.2)", rgb.to_string(**args))

    def test_no_alpha(self):
        """Test no alpha."""

        args = {"comma": True, "alpha": False, "precision": 0}

        color = "rgb(30, 100, 200, 0.2)"
        rgb = Color(color)
        self.assertEqual("rgb(30, 100, 200)", rgb.to_string(**args))

        args["comma"] = False

        color = "rgb(30 100 200 / 0.2)"
        rgb = Color(color)
        self.assertEqual("rgb(30 100 200)", rgb.to_string(**args))

        args["hex"] = True

        color = "#AA33DD22"
        rgb = Color(color)
        self.assertEqual("#aa33dd", rgb.to_string(**args))

    def test_force_alpha(self):
        """Test force alpha."""

        args = {"comma": True, "precision": 0, "alpha": True}

        color = "rgb(30, 100, 200, 100%)"
        rgb = Color(color)
        self.assertEqual("rgba(30, 100, 200, 1)", rgb.to_string(**args))

        args["comma"] = False

        color = "rgb(30 100 200 / 100%)"
        rgb = Color(color)
        self.assertEqual("rgb(30 100 200 / 1)", rgb.to_string(**args))

        args["hex"] = True

        color = "#AA33DDFF"
        rgb = Color(color)
        self.assertEqual("#aa33ddff", rgb.to_string(**args))

    def test_name_name(self):
        """Test color names."""

        args = {"names": True}

        color = "rebeccapurple"
        rgb = Color(color)
        self.assertEqual("rebeccapurple", rgb.to_string(**args))

        color = "#663399"
        rgb = Color(color)
        self.assertEqual("rebeccapurple", rgb.to_string(**args))

        args["hex"] = True
        args["names"] = False

        color = "#663399"
        rgb = Color(color)
        self.assertEqual("#663399", rgb.to_string(**args))

        args["names"] = True

        color = "#663399"
        rgb = Color(color)
        self.assertEqual("rebeccapurple", rgb.to_string(**args))

        args["alpha"] = True

        color = "#663399FF"
        rgb = Color(color)
        self.assertEqual("rebeccapurple", rgb.to_string(**args))

        color = "#66339922"
        rgb = Color(color)
        self.assertEqual("#66339922", rgb.to_string(**args))



"""Colors."""
from .srgb import SRGB
from .hsl import HSL
from .hwb import HWB
from .lab import LAB
from .lch import LCH
from ...colors import SRGB_Linear
from ...colors import HSV
from ...colors import Display_P3
from ...colors import A98_RGB
from ...colors import ProPhoto_RGB
from ...colors import Rec2020
from ...colors import XYZ
from ...colors import Color as GenericColor

SUPPORTED = (
    HSL, HWB, LAB, LCH, SRGB, SRGB_Linear, HSV,
    Display_P3, A98_RGB, ProPhoto_RGB, Rec2020, XYZ
)


class Color(GenericColor):
    """Color wrapper class."""

    CS_MAP = {obj.space(): obj for obj in SUPPORTED}



"""HWB class."""
from ._space import Space, RE_DEFAULT_MATCH
from .srgb import SRGB
from .hsl import HSL
from .hsv import HSV
from ._cylindrical import Cylindrical
from ._gamut import GamutBound
from . _range import Angle, Percent
from . import _parse as parse
from . import _convert as convert
from .. import util
import re


def srgb_to_hwb(rgb):
    """SRGB to HWB."""

    h, s, v = HSV._from_srgb(rgb)
    w = v * (100.0 - s) / 100.0
    b = 100.0 - v
    return HWB._constrain_hue(h), w, b


def hwb_to_srgb(hwb):
    """HWB to RGB."""

    h, w, b = hwb
    w /= 100.0
    b /= 100.0
    wb = w + b

    if wb > 1.0:
        return [w / wb] * 3

    return [(c * (1.0 - w - b)) + w for c in HSL._to_srgb([h, 100.0, 50.0])]


class HWB(Cylindrical, Space):
    """HWB class."""

    SPACE = "hwb"
    DEF_VALUE = "color(hwb 0 0 0 / 1)"
    CHANNEL_NAMES = frozenset(["hue", "blackness", "whiteness", "alpha"])
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    GAMUT_CHECK = "hsl"
    WHITE = convert.WHITES["D65"]

    _range = (
        GamutBound([Angle(0.0), Angle(360.0)]),
        GamutBound([Percent(0.0), Percent(100.0)]),
        GamutBound([Percent(0.0), Percent(100.0)])
    )

    def __init__(self, color=DEF_VALUE):
        """Initialize."""

        super().__init__(color)

        if isinstance(color, Space):
            self.hue, self.whiteness, self.blackness = color.convert(self.space()).coords()
            self.alpha = color.alpha
        elif isinstance(color, str):
            values = self.match(color)[0]
            if values is None:
                raise ValueError("'{}' does not appear to be a valid color".format(color))
            self.hue, self.whiteness, self.blackness, self.alpha = values
        elif isinstance(color, (list, tuple)):
            if not (3 <= len(color) <= 4):
                raise ValueError("A list of channel values should be of length 3 or 4.")
            self.hue = color[0]
            self.whiteness = color[1]
            self.blackness = color[2]
            self.alpha = 1.0 if len(color) == 3 else color[3]
        else:
            raise TypeError("Unexpected type '{}' received".format(type(color)))

    @property
    def hue(self):
        """Hue channel."""

        return self._coords[0]

    @hue.setter
    def hue(self, value):
        """Shift the hue."""

        self._coords[0] = self._handle_input(value)

    @property
    def whiteness(self):
        """Whiteness channel."""

        return self._coords[1]

    @whiteness.setter
    def whiteness(self, value):
        """Set whiteness channel."""

        self._coords[1] = self._handle_input(value)

    @property
    def blackness(self):
        """Blackness channel."""

        return self._coords[2]

    @blackness.setter
    def blackness(self, value):
        """Set blackness channel."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def null_adjust(cls, coords):
        """On color update."""

        if coords[1] + coords[2] >= 100:
            coords[0] = util.NaN
        return coords

    @classmethod
    def translate_channel(cls, channel, value):
        """Translate channel string."""

        if channel == 0:
            return parse.norm_deg_channel(value)
        elif channel in (1, 2):
            return parse.norm_float(value)
        elif channel == -1:
            return parse.norm_alpha_channel(value)
        else:
            raise ValueError("Unexpected channel index of '{}'".format(channel))

    @classmethod
    def _to_xyz(cls, hwb):
        """SRGB to XYZ."""

        return SRGB._to_xyz(cls._to_srgb(hwb))

    @classmethod
    def _from_xyz(cls, xyz):
        """XYZ to SRGB."""

        return cls._from_srgb(SRGB._from_xyz(xyz))

    @classmethod
    def _to_srgb(cls, hwb):
        """To sRGB."""

        return hwb_to_srgb(hwb)

    @classmethod
    def _from_srgb(cls, srgb):
        """From sRGB."""

        return srgb_to_hwb(srgb)

    @classmethod
    def _to_hsl(cls, hwb):
        """To HSL."""

        return HSL._from_srgb(cls._to_srgb(hwb))

    @classmethod
    def _from_hsl(cls, hsl):
        """From HSL."""

        return cls._from_srgb(HSL._to_srgb(hsl))

    @classmethod
    def _to_hsv(cls, hwb):
        """To HSV."""

        return HSV._from_srgb(cls._to_srgb(hwb))

    @classmethod
    def _from_hsv(cls, hsv):
        """From HSV."""

        return cls._from_srgb(HSV._to_srgb(hsv))



"""Color base."""
from .. import util
from . import _parse as parse
from . import _convert as convert
from . import _distance as distance
from . import _gamut as gamut
from . import _interpolate as interpolate
from . import _contrast as contrast
from . _range import Percent

# Technically this form can handle any number of channels as long as any
# extra are thrown away. We only support 6 currently. If we ever support
# colors with more channels, we can bump this.
RE_DEFAULT_MATCH = r"""(?xi)
color\(\s*
(?:({{color_space}})\s+)?
((?:{percent}|{float})(?:{space}(?:{percent}|{float})){{{{,6}}}}(?:{slash}(?:{percent}|{float}))?)
\s*\)
""".format(
    **parse.COLOR_PARTS
)


def split_channels(cls, color):
    """Split channels."""

    if color is None:
        color = ""

    channels = []
    color = color.strip()
    split = parse.RE_SLASH_SPLIT.split(color, maxsplit=1)
    alpha = None
    if len(split) > 1:
        alpha = parse.norm_alpha_channel(split[-1])
    for i, c in enumerate(parse.RE_CHAN_SPLIT.split(split[0]), 0):
        if c and i < cls.NUM_COLOR_CHANNELS:
            is_percent = isinstance(cls._range[i][0], Percent)
            channels.append(parse.norm_color_channel(c, not is_percent))
    if len(channels) < cls.NUM_COLOR_CHANNELS:
        diff = cls.NUM_COLOR_CHANNELS - len(channels)
        channels.extend([0.0] * diff)
    channels.append(alpha if alpha is not None else 1.0)
    return cls.null_adjust(channels)


class Space(contrast.Contrast, interpolate.Interpolate, distance.Distance, gamut.Gamut, convert.Convert):
    """Base color space object."""

    # Default color value (black)
    DEF_VALUE = ""
    # Color space name
    SPACE = ""
    # Number of channels
    NUM_COLOR_CHANNELS = 3
    # Channel names
    CHANNEL_NAMES = frozenset(["alpha"])
    # For matching the default form of `color(space coords+ / alpha)`.
    # Classes should define this if they want to use the default match.
    DEFAULT_MATCH = ""
    # Match pattern variable for classes to override so we can also
    # maintain the default and other alternatives.
    MATCH = ""
    # Should this color also be checked in a different color space? Only when set to a string (specifying a color space)
    # will the default gamut checking also check the specified space as well as the current.
    #
    # Gamut checking:
    #   The specified color space will be checked first followed by the original. Assuming the parent color space fits,
    #   the original should fit as well, but there are some cases when a parent color space that is slightly out of
    #   gamut, when evaluated with a threshold, may appear to be in gamut enough, but when checking the original color
    #   space, the values can be greatly out of specification (looking at you HSL).
    GAMUT_CHECK = None
    # White point
    WHITE = convert.WHITES["D50"]

    def __init__(self, color=None):
        """Initialize."""

        self.parent = None
        self._alpha = 0.0
        self._coords = [0.0] * self.NUM_COLOR_CHANNELS
        if isinstance(color, Space):
            self.parent = color.parent

    def __repr__(self):
        """Representation."""

        return 'color({} {} / {})'.format(
            self.space(),
            ' '.join([util.fmt_float(c, util.DEF_PREC) for c in util.no_nan(self.coords())]),
            util.fmt_float(util.no_nan(self.alpha), util.DEF_PREC)
        )

    __str__ = __repr__

    def _handle_input(self, value):
        """Handle numerical input."""

        if not util.is_number(value):
            raise TypeError("Value should be a number not type '{}'".format(type(value)))
        return float(value) if not util.is_nan(value) else value

    def coords(self):
        """Coordinates."""

        return self._coords[:]

    def clone(self):
        """Clone."""

        return self.new(self)

    def new(self, value):
        """Create new color in color space."""

        color = type(self)(value)
        color.parent = self.parent
        return color

    @classmethod
    def space(cls):
        """Get the color space."""

        return cls.SPACE

    @classmethod
    def white(cls):
        """Get the white color for this color space."""

        return cls.WHITE

    @property
    def alpha(self):
        """Alpha channel."""

        return self._alpha

    @alpha.setter
    def alpha(self, value):
        """Adjust alpha."""

        self._alpha = util.clamp(self._handle_input(value), 0.0, 1.0)

    def is_nan(self, name):
        """Check if the channel is NaN."""

        return util.is_nan(self.get(name))

    def set(self, name, value):  # noqa: A003
        """Set the given channel."""

        if name not in self.CHANNEL_NAMES:
            raise ValueError("'{}' is an invalid channel name".format(name))

        setattr(self, name, value)
        return self

    def get(self, name):
        """Get the given channel's value."""

        if name not in self.CHANNEL_NAMES:
            raise ValueError("'{}' is an invalid channel name".format(name))
        return getattr(self, name)

    def to_string(
        self, *, alpha=None, precision=None, fit=True, **kwargs
    ):
        """Convert to CSS 'color' string: `color(space coords+ / alpha)`."""

        if precision is None:
            precision = self.parent.PRECISION

        a = util.no_nan(self.alpha)
        alpha = alpha is not False and (alpha is True or a < 1.0)

        coords = util.no_nan(self.fit_coords() if fit else self.coords())
        template = "color({} {} {} {} / {})" if alpha else "color({} {} {} {})"
        values = [
            util.fmt_float(coords[0], precision),
            util.fmt_float(coords[1], precision),
            util.fmt_float(coords[2], precision)
        ]
        if alpha:
            values.append(util.fmt_float(a, max(precision, util.DEF_PREC)))

        return template.format(self.space(), *values)

    @classmethod
    def null_adjust(cls, coords):
        """Process coordinates and adjust any channels to null/NaN if required."""

        return coords

    @classmethod
    def translate_channel(cls, channel, value):
        """Set a non-alpha color channel."""

        raise NotImplementedError("Base 'Space' does not implement 'translate_channel' directly.")

    @classmethod
    def split_channels(cls, color):
        """Split channels."""

        raise NotImplementedError("Base 'Space' class does not implement 'translate_channel' directly.")

    @classmethod
    def match(cls, string, start=0, fullmatch=True):
        """Match a color by string."""

        m = cls.DEFAULT_MATCH.match(string, start)
        if (
            m is not None and
            (
                (m.group(1) and m.group(1).lower() == cls.space())
            ) and (not fullmatch or m.end(0) == len(string))
        ):
            return split_channels(cls, m.group(2)), m.end(0)
        return None, None



"""SRGB color class."""
from ._space import Space
from ._space import RE_DEFAULT_MATCH
from ._gamut import GamutBound
from .xyz import XYZ
from . import _parse as parse
from . import _convert as convert
from .. import util
import re
import math


def lin_srgb_to_xyz(rgb):
    """
    Convert an array of linear-light sRGB values to CIE XYZ using sRGB's own white.

    D65 (no chromatic adaptation)
    """

    m = [
        [0.41239079926595934, 0.357584339383878, 0.1804807884018343],
        [0.21263900587151027, 0.715168678767756, 0.07219231536073371],
        [0.01933081871559182, 0.11919477979462598, 0.9505321522496607]
    ]

    return util.dot(m, rgb)


def xyz_to_lin_srgb(xyz):
    """Convert XYZ to linear-light sRGB."""

    m = [
        [3.2409699419045226, -1.537383177570094, -0.4986107602930034],
        [-0.9692436362808796, 1.8759675015077202, 0.04155505740717559],
        [0.05563007969699366, -0.20397695888897652, 1.0569715142428786]
    ]

    return util.dot(m, xyz)


def lin_srgb(rgb):
    """
    Convert an array of sRGB values in the range 0.0 - 1.0 to linear light (un-corrected) form.

    https://en.wikipedia.org/wiki/SRGB
    """

    result = []
    for i in rgb:
        # Mirror linear nature of algorithm on the negative axis
        abs_i = abs(i)
        if abs_i < 0.04045:
            result.append(i / 12.92)
        else:
            result.append(math.copysign(math.pow((abs_i + 0.055) / 1.055, 2.4), i))
    return result


def gam_srgb(rgb):
    """
    Convert an array of linear-light sRGB values in the range 0.0-1.0 to gamma corrected form.

    https://en.wikipedia.org/wiki/SRGB
    """

    result = []
    for i in rgb:
        # Mirror linear nature of algorithm on the negative axis
        abs_i = abs(i)
        if abs_i > 0.0031308:
            result.append(math.copysign((1.055 * math.pow(abs_i, 1 / 2.4) - 0.055), i))
        else:
            result.append(12.92 * i)
    return result


class SRGB(Space):
    """SRGB class."""

    SPACE = "srgb"
    # In addition to the current gamut, check HSL as it is much more sensitive to small
    # gamut changes. This is mainly for a better user experience. Colors will still be
    # mapped/clipped in the current space, unless specified otherwise.
    GAMUT_CHECK = "hsl"
    DEF_VALUE = "color(srgb 0 0 0 / 1)"
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    CHANNEL_NAMES = frozenset(["red", "green", "blue", "alpha"])
    WHITE = convert.WHITES["D65"]

    _range = (
        GamutBound([0.0, 1.0]),
        GamutBound([0.0, 1.0]),
        GamutBound([0.0, 1.0])
    )

    def __init__(self, color=DEF_VALUE):
        """Initialize."""

        super().__init__(color)

        if isinstance(color, Space):
            self.red, self.green, self.blue = color.convert(self.space()).coords()
            self.alpha = color.alpha
        elif isinstance(color, str):
            values = self.match(color)[0]
            if values is None:
                raise ValueError("'{}' does not appear to be a valid color".format(color))
            self.red, self.green, self.blue, self.alpha = values
        elif isinstance(color, (list, tuple)):
            if not (3 <= len(color) <= 4):
                raise ValueError("A list of channel values should be of length 3 or 4.")
            self.red = color[0]
            self.green = color[1]
            self.blue = color[2]
            self.alpha = 1.0 if len(color) == 3 else color[3]
        else:
            raise TypeError("Unexpected type '{}' received".format(type(color)))

    @property
    def red(self):
        """Adjust red."""

        return self._coords[0]

    @red.setter
    def red(self, value):
        """Adjust red."""

        self._coords[0] = self._handle_input(value)

    @property
    def green(self):
        """Adjust green."""

        return self._coords[1]

    @green.setter
    def green(self, value):
        """Adjust green."""

        self._coords[1] = self._handle_input(value)

    @property
    def blue(self):
        """Adjust blue."""

        return self._coords[2]

    @blue.setter
    def blue(self, value):
        """Adjust blue."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def translate_channel(cls, channel, value):
        """Translate channel string."""

        if 0 <= channel <= 2:
            return parse.norm_float(value)
        elif channel == -1:
            return parse.norm_alpha_channel(value)
        else:
            raise ValueError("Unexpected channel index of '{}'".format(channel))

    @classmethod
    def _to_xyz(cls, rgb):
        """SRGB to XYZ."""

        return cls._chromatic_adaption(cls.white(), XYZ.white(), lin_srgb_to_xyz(lin_srgb(rgb)))

    @classmethod
    def _from_xyz(cls, xyz):
        """XYZ to SRGB."""

        return gam_srgb(xyz_to_lin_srgb(cls._chromatic_adaption(XYZ.white(), cls.white(), xyz)))



"""LAB class."""
import re
from ...colors import lab as generic
from ...colors import _parse as parse
from ... import util


class LAB(generic.LAB):
    """LAB class."""

    DEF_VALUE = "lab(0% 0 0 / 1)"
    START = re.compile(r'(?i)\blab\(')
    MATCH = re.compile(
        r"""(?xi)
        (?:
            \blab\(\s*
            (?:
                # Space separated format
                {percent}{space}{float}{space}{float}(?:{slash}(?:{percent}|{float}))? |
                # comma separated format
                {percent}{comma}{float}{comma}{float}(?:{comma}(?:{percent}|{float}))?
            )
            \s*\)
        )
        """.format(**parse.COLOR_PARTS)
    )

    def __init__(self, color=DEF_VALUE):
        """Initialize."""

        super().__init__(color)

    def to_string(
        self, *, alpha=None, precision=None, fit=True, **kwargs
    ):
        """Convert to CSS."""

        if precision is None:
            precision = self.parent.PRECISION

        options = kwargs
        if options.get("color"):
            return super().to_string(alpha=alpha, precision=precision, fit=fit, **kwargs)

        a = util.no_nan(self.alpha)
        alpha = alpha is not False and (alpha is True or a < 1.0)
        coords = util.no_nan(self.fit_coords() if fit else self.coords())

        if alpha:
            template = "lab({}%, {}, {}, {})" if options.get("comma") else "lab({}% {} {} / {})"
            return template.format(
                util.fmt_float(coords[0], precision),
                util.fmt_float(coords[1], precision),
                util.fmt_float(coords[2], precision),
                util.fmt_float(a, max(util.DEF_PREC, precision))
            )
        else:
            template = "lab({}%, {}, {})" if options.get("comma") else "lab({}% {} {})"
            return template.format(
                util.fmt_float(coords[0], precision),
                util.fmt_float(coords[1], precision),
                util.fmt_float(coords[2], precision)
            )

    @classmethod
    def translate_channel(cls, channel, value):
        """Translate channel string."""

        if channel == 0:
            return parse.norm_lab_lightness(value)
        elif channel in (1, 2):
            return parse.norm_float(value)
        elif channel == -1:
            return parse.norm_alpha_channel(value)
        else:
            raise ValueError("Unexpected channel index of '{}'".format(channel))

    @classmethod
    def split_channels(cls, color):
        """Split channels."""

        start = 4
        channels = []
        for i, c in enumerate(parse.RE_CHAN_SPLIT.split(color[start:-1].strip()), 0):
            if i <= 2:
                channels.append(cls.translate_channel(i, c))
            else:
                channels.append(cls.translate_channel(-1, c))
        if len(channels) == 3:
            channels.append(1.0)
        return cls.null_adjust(channels)

    @classmethod
    def match(cls, string, start=0, fullmatch=True):
        """Match a CSS color string."""

        channels, end = super().match(string, start, fullmatch)
        if channels is not None:
            return channels, end
        m = cls.MATCH.match(string, start)
        if m is not None and (not fullmatch or m.end(0) == len(string)):
            return cls.split_channels(string[m.start(0):m.end(0)]), m.end(0)
        return None, None
